/*  CSS format as required by Debian (GTK 3.14.5)
	
	https://developer.gnome.org/gtk3/stable/chap-css-properties.html
	https://developer.gnome.org/gtk3/stable/gtk-migrating-GtkStyleContext-css.html
	pkg-config --modversion gtk+-3.0                                                   */

*
{
	/* see wx+/grid.cpp: spacing wouldn't hurt, but let's be consistent */
	-GtkScrolledWindow-scrollbar-spacing: 0;
}

GtkButton
{
	padding: 2px; /*remove excessive inner border from bitmap buttons*/
	/*	min-width:  0;  => Debian: Error code 3: Gtk3Styles.css:13:10'min-width' is not a valid property name [gtk_css_provider_load_from_path]
		min-height: 0;      */
}
