/*  CSS format as required by CentOS (GTK 3.22.30)

	https://developer.gnome.org/gtk3/stable/chap-css-properties.html
	https://developer.gnome.org/gtk3/stable/gtk-migrating-GtkStyleContext-css.html
	pkg-config --modversion gtk+-3.0                                                      */

*
{
	/* see wx+/grid.cpp: spacing wouldn't hurt, but let's be consistent */
	-GtkScrolledWindow-scrollbar-spacing: 0;
}

button
{
	padding: 2px; /*remove excessive inner border from bitmap buttons*/
	min-width:  0;
	min-height: 0;
	/*border-radius: 5px;*/
}
