style "no-inner-border"
{
	GtkButton::inner-border = {0, 0, 0, 0} /*remove excessive borders on Gnome*/
	/*GtkButton::focus-padding = 0        => keep default: minor difference + looks better on KDE */
}

class "GtkButton" style "no-inner-border"


style "no-scrollbar-spacing"
{
	/* see wx+/grid.cpp: implementation assumes no spacing! */
	GtkScrolledWindow::scrollbar-spacing = 0
}

class "GtkScrolledWindow" style "no-scrollbar-spacing"
